#!/usr/bin/env python3
import timeit

import numpy as np

from extraction.core.functions.custom.localisation import nuc_est_conv

image_size = 100
mask_size = 20
image = np.reshape(
    np.random.randint(100, size=image_size ** 2), (image_size, image_size)
)
mask = np.pad(np.ones((mask_size, mask_size)), (image_size - mask_size) // 2).astype(
    bool
)

wrapped = nuc_est_conv(mask, image)
# timeit.timeit(wrapped, number=1000)
# import cProfile
# import pstats

# profile = cProfile.Profile()
# profile.runcall(wrapped)
# ps = pstats.Stats(profile)
# ps.print_stats()



"""
Classes to handle multidimensional images, both remotely and local.
"""
import itertools
import logging

import h5py
import numpy as np
from pathlib import Path

from tqdm import tqdm
import cv2

from aliby.io.matlab import matObject
from agora.io.utils import Cache, imread, get_store_path

logger = logging.getLogger(__name__)


def parse_local_fs(pos_dir, tp=None):
    """
    Local file structure:
    - pos_dir
        -- exptID_{timepointID}_{ChannelID}_{z_position_id}.png

    :param pos_dirs:
    :return: Image_mapper
    """
    pos_dir = Path(pos_dir)

    img_mapper = dict()

    def channel_idx(img_name):
        return img_name.stem.split("_")[-2]

    def tp_idx(img_name):
        return int(img_name.stem.split("_")[-3]) - 1

    def z_idx(img_name):
        return img_name.stem.split("_")[-1]

    if tp is not None:
        img_list = [img for img in pos_dir.iterdir() if tp_idx(img) in tp]
    else:
        img_list = [img for img in pos_dir.iterdir()]

    for tp, group in itertools.groupby(sorted(img_list, key=tp_idx), key=tp_idx):
        img_mapper[int(tp)] = {
            channel: {i: item for i, item in enumerate(sorted(grp, key=z_idx))}
            for channel, grp in itertools.groupby(
                sorted(group, key=channel_idx), key=channel_idx
            )
        }
    return img_mapper


class Timelapse:
    """
    Timelapse class contains the specifics of one position.
    """

    def __init__(self):
        self._id = None
        self._name = None
        self._channels = []
        self._size_c = 0
        self._size_t = 0
        self._size_x = 0
        self._size_y = 0
        self._size_z = 0
        self.image_cache = None
        self.annotation = None

    def __repr__(self):
        return self.name

    def full_mask(self):
        return np.full(self.shape, False)

    def __getitem__(self, item):
        cached = self.image_cache[item]
        # Check if there are missing values, if so reload
        # TODO only reload missing
        mask = np.isnan(cached)
        if np.any(mask):
            full = self.load_fn(item)
            shape = self.image_cache[
                item
            ].shape  # TODO speed this up by  recognising the shape from the item
            self.image_cache[item] = np.reshape(full, shape)
            return full
        return cached

    def get_hypercube(self):
        pass

    def load_fn(self, item):
        """
        The hypercube is ordered as: C, T, X, Y, Z
        :param item:
        :return:
        """

        def parse_slice(s):
            step = s.step if s.step is not None else 1
            if s.start is None and s.stop is None:
                return None
            elif s.start is None and s.stop is not None:
                return range(0, s.stop, step)
            elif s.start is not None and s.stop is None:
                return [s.start]
            else:  # both s.start and s.stop are not None
                return range(s.start, s.stop, step)

        def parse_subitem(subitem, kw):
            if isinstance(subitem, (int, float)):
                res = [int(subitem)]
            elif isinstance(subitem, list) or isinstance(subitem, tuple):
                res = list(subitem)
            elif isinstance(subitem, slice):
                res = parse_slice(subitem)
            else:
                res = subitem
                # raise ValueError(f"Cannot parse slice {kw}: {subitem}")

            if kw in ["x", "y"]:
                # Need exactly two values
                if res is not None:
                    if len(res) < 2:
                        # An int was passed, assume it was
                        res = [res[0], self.size_x]
                    elif len(res) > 2:
                        res = [res[0], res[-1] + 1]
            return res

        if isinstance(item, int):
            return self.get_hypercube(
                x=None, y=None, z_positions=None, channels=[item], timepoints=None
            )
        elif isinstance(item, slice):
            return self.get_hypercube(channels=parse_slice(item))
        keywords = ["channels", "timepoints", "x", "y", "z_positions"]
        kwargs = dict()
        for kw, subitem in zip(keywords, item):
            kwargs[kw] = parse_subitem(subitem, kw)
        return self.get_hypercube(**kwargs)

    @property
    def shape(self):
        return (self.size_c, self.size_t, self.size_x, self.size_y, self.size_z)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def size_z(self):
        return self._size_z

    @property
    def size_c(self):
        return self._size_c

    @property
    def size_t(self):
        return self._size_t

    @property
    def size_x(self):
        return self._size_x

    @property
    def size_y(self):
        return self._size_y

    @property
    def channels(self):
        return self._channels

    def get_channel_index(self, channel):
        return self.channels.index(channel)


def load_annotation(filepath: Path):
    try:
        return matObject(filepath)
    except Exception as e:
        raise (
            "Could not load annotation file. \n"
            "Non MATLAB files currently unsupported"
        ) from e


class TimelapseOMERO(Timelapse):
    """
    Connected to an Image object which handles database I/O.
    """

    def __init__(self, image, annotation, cache, **kwargs):
        super(TimelapseOMERO, self).__init__()
        self.image = image
        # Pre-load pixels
        self.pixels = self.image.getPrimaryPixels()
        self._id = self.image.getId()
        self._name = self.image.getName()
        self._size_x = self.image.getSizeX()
        self._size_y = self.image.getSizeY()
        self._size_z = self.image.getSizeZ()
        self._size_c = self.image.getSizeC()
        self._size_t = self.image.getSizeT()
        self._channels = self.image.getChannelLabels()
        # Check whether there are file annotations for this position
        if annotation is not None:
            self.annotation = load_annotation(annotation)
        # Get an HDF5 dataset to use as a cache.
        compression = kwargs.get("compression", None)
        self.image_cache = cache.require_dataset(
            self.name,
            self.shape,
            dtype=np.float16,
            fillvalue=np.nan,
            compression=compression,
        )

    def get_hypercube(
        self, x=None, y=None, z_positions=None, channels=None, timepoints=None
    ):
        if x is None and y is None:
            tile = None  # Get full plane
        elif x is None:
            ymin, ymax = y
            tile = (None, ymin, None, ymax - ymin)
        elif y is None:
            xmin, xmax = x
            tile = (xmin, None, xmax - xmin, None)
        else:
            xmin, xmax = x
            ymin, ymax = y
            tile = (xmin, ymin, xmax - xmin, ymax - ymin)

        if z_positions is None:
            z_positions = range(self.size_z)
        if channels is None:
            channels = range(self.size_c)
        if timepoints is None:
            timepoints = range(self.size_t)

        z_positions = z_positions or [0]
        channels = channels or [0]
        timepoints = timepoints or [0]

        zcttile_list = [
            (z, c, t, tile)
            for z, c, t in itertools.product(z_positions, channels, timepoints)
        ]
        planes = list(self.pixels.getTiles(zcttile_list))
        order = (
            len(z_positions),
            len(channels),
            len(timepoints),
            planes[0].shape[-2],
            planes[0].shape[-1],
        )
        result = np.stack([x for x in planes]).reshape(order)
        # Set to C, T, X, Y, Z order
        result = np.moveaxis(result, -1, -2)
        return np.moveaxis(result, 0, -1)

    def cache_set(self, save_dir, timepoints, expt_name, quiet=True):
        # TODO deprecate when this is default
        pos_dir = save_dir / self.name
        if not pos_dir.exists():
            pos_dir.mkdir()
        for tp in tqdm(timepoints, desc=self.name):
            for channel in tqdm(self.channels, disable=quiet):
                for z_pos in tqdm(range(self.size_z), disable=quiet):
                    ch_id = self.get_channel_index(channel)
                    image = self.get_hypercube(
                        x=None,
                        y=None,
                        channels=[ch_id],
                        z_positions=[z_pos],
                        timepoints=[tp],
                    )
                    im_name = "{}_{:06d}_{}_{:03d}.png".format(
                        expt_name, tp + 1, channel, z_pos + 1
                    )
                    cv2.imwrite(str(pos_dir / im_name), np.squeeze(image))
        # TODO update positions table to get the number of timepoints?
        return list(itertools.product([self.name], timepoints))

    def run(self, keys, store, save_dir="./", **kwargs):
        """
        Parse file structure and get images for the timepoints in keys.
        """
        save_dir = Path(save_dir)
        if keys is None:
            # TODO save final metadata
            return None
        store = save_dir / store
        # A position specific store
        store = store.with_name(self.name + store.name)
        # Create store if it does not exist
        if not store.exists():
            # The first run, add metadata to the store
            with h5py.File(store, "w") as pos_store:
                # TODO Add metadata to the store.
                pass
        # TODO check how sensible the keys are with what is available
        #   if some of the keys don't make sense, log a warning and remove
        #   them so that the next steps of the pipeline make sense
        return keys

    def clear_cache(self):
        self.image_cache.clear()


class TimelapseLocal(Timelapse):
    def __init__(
        self, position, root_dir, finished=True, annotation=None, cache=None, **kwargs
    ):
        """
        Linked to a local directory containing the images for one position
        in an experiment.
        Can be a still running experiment or a finished one.

        :param position: Name of the position
        :param root_dir: Root directory
        :param finished: Whether the experiment has finished running or the
        class will be used as part of a pipeline, mostly with calls to `run`
        """
        super(TimelapseLocal, self).__init__()
        self.pos_dir = Path(root_dir) / position
        assert self.pos_dir.exists()
        self._id = position
        self._name = position
        if finished:
            self.image_mapper = parse_local_fs(self.pos_dir)
            self._update_metadata()
        else:
            self.image_mapper = dict()
        self.annotation = None
        # Check whether there are file annotations for this position
        if annotation is not None:
            self.annotation = load_annotation(annotation)
        compression = kwargs.get("compression", None)
        self.image_cache = cache.require_dataset(
            self.name,
            self.shape,
            dtype=np.float16,
            fillvalue=np.nan,
            compression=compression,
        )

    def _update_metadata(self):
        self._size_t = len(self.image_mapper)
        # Todo: if cy5 is the first one it causes issues with getting x, y
        #   hence the sorted but it's not very robust
        self._channels = sorted(
            list(set.union(*[set(tp.keys()) for tp in self.image_mapper.values()]))
        )
        self._size_c = len(self._channels)
        # Todo: refactor so we don't rely on there being any images at all
        self._size_z = max([len(self.image_mapper[0][ch]) for ch in self._channels])
        single_img = self.get_hypercube(
            x=None, y=None, z_positions=None, channels=[0], timepoints=[0]
        )
        self._size_x = single_img.shape[2]
        self._size_y = single_img.shape[3]

    def get_hypercube(
        self, x=None, y=None, z_positions=None, channels=None, timepoints=None
    ):
        xmin, xmax = x if x is not None else (None, None)
        ymin, ymax = y if y is not None else (None, None)

        if z_positions is None:
            z_positions = range(self.size_z)
        if channels is None:
            channels = range(self.size_c)
        if timepoints is None:
            timepoints = range(self.size_t)

        def z_pos_getter(z_positions, ch_id, t):
            default = np.zeros((self.size_x, self.size_y))
            names = [
                self.image_mapper[t][self.channels[ch_id]].get(i, None)
                for i in z_positions
            ]
            res = [imread(name) if name is not None else default for name in names]
            return res

        # nested list of images in C, T, X, Y, Z order
        ctxyz = []
        for ch_id in channels:
            txyz = []
            for t in timepoints:
                xyz = z_pos_getter(z_positions, ch_id, t)
                txyz.append(np.dstack(list(xyz))[xmin:xmax, ymin:ymax])
            ctxyz.append(np.stack(txyz))
        return np.stack(ctxyz)

    def clear_cache(self):
        self.image_cache.clear()

    def run(self, keys, store, save_dir="./", **kwargs):
        """
        Parse file structure and get images for the time points in keys.
        """
        if keys is None:
            return None
        elif isinstance(keys, int):
            keys = [keys]
        self.image_mapper.update(parse_local_fs(self.pos_dir, tp=keys))
        self._update_metadata()
        # Create store if it does not exist
        store = get_store_path(save_dir, store, self.name)
        if not store.exists():
            # The first run, add metadata to the store
            with h5py.File(store, "w") as pos_store:
                # TODO Add metadata to the store.
                pass
        # TODO check how sensible the keys are with what is available
        #   if some of the keys don't make sense, log a warning and remove
        #   them so that the next steps of the pipeline make sense
        return keys



"""
Segment/segmented pipelines.
Includes splitting the image into traps/parts,
cell segmentation, nucleus segmentation

Standard order is (T, C, Z, Y, X)
"""
import warnings
from functools import lru_cache
from pathlib import PosixPath
from typing import Union

import dask.array as da
import h5py
import numpy as np
from agora.abc import ParametersABC, ProcessABC
from agora.io.writer import load_attributes
from skimage.registration import phase_cross_correlation

from aliby.io.image import Image, ImageLocal
from aliby.tile.traps import segment_traps


class Trap:
    """
    Stores a trap's location and size.
    Allows checks to see if the trap should be padded.
    Can export the trap either in OMERO or numpy formats.
    """

    def __init__(self, centre, parent, size, max_size):
        self.centre = centre
        self.parent = parent  # used to access drifts
        self.size = size
        self.half_size = size // 2
        self.max_size = max_size

    def padding_required(self, tp):
        """
        Check if we need to pad the trap image for this time point.

        Parameters
        ----------
        tp: integer
            Index for a time point
        """
        try:
            assert all(self.at_time(tp) - self.half_size >= 0)
            assert all(self.at_time(tp) + self.half_size <= self.max_size)
            # return False
        except AssertionError:
            return True
        return False

    def at_time(self, tp):
        """
        Return trap centre at time tp by applying drifts

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        trap_centre:
        """
        drifts = self.parent.drifts
        trap_centre = self.centre - np.sum(drifts[: tp + 1], axis=0)
        return trap_centre

    def as_tile(self, tp):
        """
        Return trap in the OMERO tile format of x, y, w, h
        where x, y are at the bottom left corner of the tile
        and w and h are the tile width and height.

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        x: int
            x-coordinate of bottom left corner of tile
        y: int
            y-coordinate of bottom left corner of tile
        w: int
            Width of tile
        h: int
            Height of tile
        """
        x, y = self.at_time(tp)
        # tile bottom corner
        x = int(x - self.half_size)
        y = int(y - self.half_size)
        return x, y, self.size, self.size

    def as_range(self, tp):
        """
        Return trap in a range format: two slice objects that can
        be used in arrays

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        A slice of x coordinates from left to right
        A slice of y coordinates from top to bottom
        """
        x, y, w, h = self.as_tile(tp)
        return slice(x, x + w), slice(y, y + h)


class TrapLocations:
    """
    Stores each trap as an instance of Trap.
    Traps can be iterated.
    """

    def __init__(
        self,
        initial_location: np.array,
        tile_size: int = None,
        max_size: int = 1200,
        drifts: np.array = None,
    ):
        if drifts is None:
            drifts = []
        self.tile_size = tile_size
        self.max_size = max_size
        self.initial_location = initial_location
        self.traps = [
            Trap(centre, self, tile_size or max_size, max_size)
            for centre in initial_location
        ]
        self.drifts = drifts

    def __len__(self):
        return len(self.traps)

    def __iter__(self):
        yield from self.traps

    @property
    def shape(self):
        """
        Returns no of traps and no of drifts
        """
        return len(self.traps), len(self.drifts)

    def padding_required(self, tp):
        """
        Check if any traps need padding

        Parameters
        ----------
        tp: integer
            An index for a time point
        """
        return any([trap.padding_required(tp) for trap in self.traps])

    def to_dict(self, tp):
        """
        Export inital locations, tile_size, max_size, and drifts
        as a dictionary

        Parameters
        ----------
        tp: integer
            An index for a time point
        """
        res = dict()
        if tp == 0:
            res["trap_locations"] = self.initial_location
            res["attrs/tile_size"] = self.tile_size
            res["attrs/max_size"] = self.max_size
        res["drifts"] = np.expand_dims(self.drifts[tp], axis=0)
        return res

    @classmethod
    def from_tiler_init(
        cls, initial_location, tile_size: int = None, max_size: int = 1200
    ):
        """
        Instantiate class from an instance of the Tiler class
        """
        return cls(initial_location, tile_size, max_size, drifts=[])

    @classmethod
    def read_hdf5(cls, file):
        """
        Instantiate class from a hdf5 file
        """
        with h5py.File(file, "r") as hfile:
            trap_info = hfile["trap_info"]
            initial_locations = trap_info["trap_locations"][()]
            drifts = trap_info["drifts"][()].tolist()
            max_size = trap_info.attrs["max_size"]
            tile_size = trap_info.attrs["tile_size"]
        trap_locs = cls(initial_locations, tile_size, max_size=max_size)
        trap_locs.drifts = drifts
        return trap_locs


class TilerParameters(ParametersABC):
    _defaults = {"tile_size": 117, "ref_channel": "Brightfield", "ref_z": 0}


class Tiler(ProcessABC):
    """
    Remote Timelapse Tiler.

    Finds traps and re-registers images if there is any drifting.
    Fetches images from a server.
    """

    def __init__(
        self,
        image: da.core.Array,
        metadata: dict,
        parameters: TilerParameters,
        trap_locs=None,
    ):
        """
        Initialise Tiler

        Parameters
        ----------
        image: an instance of Image
        metadata: dictionary
        parameters: an instance of TilerPameters
        trap_locs: (optional)
        """
        super().__init__(parameters)
        self.image = image
        self._metadata = metadata
        self.channels = metadata["channels"]
        assert (
            parameters.ref_channel in self.channels
        ), "Reference channel not in the available channels"
        self.ref_channel = self.get_channel_index(parameters.ref_channel)
        self.trap_locs = trap_locs
        try:
            self.z_perchannel = {
                ch: metadata["zsectioning/nsections"] if zsect else 1
                for zsect, ch in zip(
                    metadata["channels"], metadata["channels/zsect"]
                )
            }
        except Exception as e:
            print(f"Warning:Tiler: No z_perchannel data: {e}")

        self.tile_size = self.tile_size or min(self.image.shape[-2:])

    @classmethod
    def from_image(cls, image: Image, parameters: TilerParameters):
        """
        Instantiate Tiler from an Image instance

        Parameters
        ----------
        image: an instance of Image
        parameters: an instance of TilerPameters
        """
        return cls(image.data, image.metadata, parameters)

    @classmethod
    def from_hdf5(
        cls,
        image: Union[Image, ImageLocal],
        filepath: Union[str, PosixPath],
        parameters: TilerParameters = None,
    ):
        """
        Instantiate Tiler from hdf5 files

        Parameters
        ----------
        image: an instance of Image
        filepath: Path instance
            Path to a directory of h5 files
        parameters: an instance of TileParameters (optional)
        """
        trap_locs = TrapLocations.read_hdf5(filepath)
        metadata = load_attributes(filepath)
        metadata["channels"] = image.metadata["channels"]
        # metadata["zsectioning/nsections"] = image.metadata["zsectioning/nsections"]
        # metadata["channels/zsect"] = image.metadata["channels/zsect"]
        if parameters is None:
            parameters = TilerParameters.default()
        tiler = cls(
            image.data,
            metadata,
            parameters,
            trap_locs=trap_locs,
        )
        if hasattr(trap_locs, "drifts"):
            tiler.n_processed = len(trap_locs.drifts)
        return tiler

    @lru_cache(maxsize=2)
    def get_tc(self, t, c):
        """
            Load image using dask.
            Assumes the image is arranged as
                no of time points
                no of channels
                no of z stacks
                no of pixels in y direction
                no of pixels in z direction

            Parameters
            ----------
            t: integer
                An index for a time point
            c: integer
                An index for a channel

        def _initialise_traps(self, tile_size: int):
            -------
            full: an array of images
        """
        full = self.image[t, c].compute(scheduler="synchronous")

        return full

    @property
    def shape(self):
        """
        Returns properties of the time-lapse experiment
            no of channels
            no of time points
            no of z stacks
            no of pixels in y direction
            no of pixels in z direction
        """
        c, t, z, y, x = self.image.shape
        return (c, t, x, y, z)

    @property
    def n_processed(self):
        """
        Returns the number of images that have been processed
        """
        if not hasattr(self, "_n_processed"):
            self._n_processed = 0
        return self._n_processed

    @n_processed.setter
    def n_processed(self, value):
        self._n_processed = value

    @property
    def n_traps(self):
        """
        Returns number of traps
        """
        return len(self.trap_locs)

    def initialise_traps(self, tile_size: int = None):
        """
        Find initial trap positions if they have not been initialised.
        Removes all those that are too close to the edge so no padding
        is necessary.

        Parameters
        ----------
        tile_size: integer
            The size of a tile
        """
        initial_image = self.image[0, self.ref_channel, self.ref_z]
        if tile_size:
            half_tile = tile_size // 2
            # max_size is the minimal number of x or y pixels
            max_size = min(self.image.shape[-2:])
            # first time point, reference channel, reference z-position
            # find the traps
            trap_locs = segment_traps(initial_image, tile_size)
            # keep only traps that are not near an edge
            trap_locs = [
                [x, y]
                for x, y in trap_locs
                if half_tile < x < max_size - half_tile
                and half_tile < y < max_size - half_tile
            ]
            # store traps in an instance of TrapLocations
            self.trap_locs = TrapLocations.from_tiler_init(
                trap_locs, tile_size
            )
        else:
            yx_shape = self.image.shape[-2:]
            trap_locs = [[x // 2 for x in yx_shape]]
            self.trap_locs = TrapLocations.from_tiler_init(
                trap_locs, max_size=min(yx_shape)
            )

    def find_drift(self, tp):
        """
        Find any translational drifts between two images at consecutive
        time points using cross correlation.

        Arguments
        ---------
        tp: integer
            Index for a time point
        """
        prev_tp = max(0, tp - 1)
        # cross-correlate
        drift, error, _ = phase_cross_correlation(
            self.image[prev_tp, self.ref_channel, self.ref_z],
            self.image[tp, self.ref_channel, self.ref_z],
        )
        # store drift
        if 0 < tp < len(self.trap_locs.drifts):
            self.trap_locs.drifts[tp] = drift.tolist()
        else:
            self.trap_locs.drifts.append(drift.tolist())

    def get_tp_data(self, tp, c):
        """
        Returns all traps corrected for drift.

        Parameters
        ----------
        tp: integer
            An index for a time point
        c: integer
            An index for a channel
        """
        traps = []
        # get OMERO image
        full = self.get_tc(tp, c)
        for trap in self.trap_locs:
            # pad trap if necessary
            ndtrap = self.ifoob_pad(full, trap.as_range(tp))
            traps.append(ndtrap)
        return np.stack(traps)

    def get_trap_data(self, trap_id, tp, c):
        """
        Returns a particular trap corrected for drift and padding

        Parameters
        ----------
        trap_id: integer
            Number of trap
        tp: integer
            Index of time points
        c: integer
            Index of channel

        Returns
        -------
        ndtrap: array
            An array of (x, y) arrays, one for each z stack
        """
        full = self.get_tc(tp, c)
        trap = self.trap_locs.traps[trap_id]
        ndtrap = self.ifoob_pad(full, trap.as_range(tp))
        return ndtrap

    def run_tp(self, tp):
        """
        Find traps if they have not yet been found.
        Determine any translational drift of the current image from the
        previous one.

        Arguments
        ---------
        tp: integer
            The time point to tile.
        """
        # assert tp >= self.n_processed, "Time point already processed"
        # TODO check contiguity?
        if self.n_processed == 0 or not hasattr(self.trap_locs, "drifts"):
            self.initialise_traps(self.tile_size)
        if hasattr(self.trap_locs, "drifts"):
            drift_len = len(self.trap_locs.drifts)
            if self.n_processed != drift_len:
                raise Exception("Tiler:n_processed and ndrifts don't match")
                self.n_processed = drift_len
        # determine drift
        self.find_drift(tp)
        # update n_processed
        self.n_processed = tp + 1
        # return result for writer
        return self.trap_locs.to_dict(tp)

    def run(self, time_dim=None):
        """
        Tile all time points in an experiment at once.
        """
        if time_dim is None:
            time_dim = 0
        for t in range(self.image.shape[time_dim]):
            self.run_tp(t)

        return None

    # The next set of functions are necessary for the extraction object
    def get_traps_timepoint(self, tp, tile_size=None, channels=None, z=None):
        # FIXME we currently ignore the tile size
        # FIXME can we ignore z(always  give)
        if isinstance(channels, str):
            channels = [channels]
        res = []
        for c in channels:
            val = self.get_tp_data(tp, c)[:, z]  # Only return requested z
            # positions
            # Starts at traps, z, y, x
            # Turn to Trap, C, T, X, Y, Z order
            val = val.swapaxes(1, 3).swapaxes(1, 2)
            val = np.expand_dims(val, axis=1)
            res.append(val)
        return np.stack(res, axis=1)

    def get_channel_index(self, item):
        """
        Find index for channel

        Parameters
        ----------
        item: string
            The channel
        """
        for i, ch in enumerate(self.channels):
            if item in ch:
                return i

    def get_position_annotation(self):
        # TODO required for matlab support
        return None

    @staticmethod
    def ifoob_pad(full, slices):  # TODO Remove when inheriting TilerABC
        """
        Returns the slices padded if it is out of bounds.

        Parameters
        ----------
        full: array
            Slice of OMERO image (zstacks, x, y) - the entire position
            with zstacks as first axis
        slices: tuple of two slices
            Delineates indiceds for the x- and y- ranges of the tile.

        Returns
        -------
        trap: array
            A tile with all z stacks for the given slices.
            If some padding is needed, the median of the image is used.
            If much padding is needed, a tile of NaN is returned.
        """
        # number of pixels in the y direction
        max_size = full.shape[-1]
        # ignore parts of the tile outside of the image
        y, x = [slice(max(0, s.start), min(max_size, s.stop)) for s in slices]
        # get the tile including all z stacks
        trap = full[:, y, x]
        # find extent of padding needed in x and y
        padding = np.array(
            [(-min(0, s.start), -min(0, max_size - s.stop)) for s in slices]
        )
        if padding.any():
            tile_size = slices[0].stop - slices[0].start
            if (padding > tile_size / 4).any():
                # too much of the tile is outside of the image
                # fill with NaN
                trap = np.full((full.shape[0], tile_size, tile_size), np.nan)
            else:
                # pad tile with median value of trap image
                trap = np.pad(trap, [[0, 0]] + padding.tolist(), "median")
        return trap



"""
A set of utilities for dealing with ALCATRAS traps
"""

from copy import copy

import numpy as np
import matplotlib.colors as colors
from tqdm import tqdm

from skimage import transform, feature
from skimage.filters.rank import entropy
from skimage.filters import threshold_otsu
from skimage.segmentation import clear_border
from skimage.measure import label, regionprops
from skimage.morphology import disk, closing, square
from skimage.registration import phase_cross_correlation
from skimage.util import img_as_ubyte


def half_floor(x, tile_size):
    return x - tile_size // 2


def half_ceil(x, tile_size):
    return x + -(tile_size // -2)


def segment_traps(
    image,
    tile_size,
    downscale=0.4,
    disk_radius_frac=0.01,
    square_size=3,
    min_frac_tilesize=0.2,
    max_frac_tilesize=0.8,
    **identify_traps_kwargs,
):
    """
    Uses an entropy filter and Otsu thresholding to find a trap template,
    which is then passed to identify_trap_locations.

    The hyperparameters have not been optimised.

    Parameters
    ----------
    image: 2D array
    tile_size: integer
        Size of the tile
    downscale: float (optional)
        Fraction by which to shrink image
    disk_radius_frac: float (optional)
        Radius of disk using in the entropy filter
    square_size: integer (optional)
        Parameter for a morphological closing applied to thresholded
        image
    min_frac_tilesize: float (optional)
    max_frac_tilesize: float (optional)
        Used to determine bounds on the major axis length of regions
        suspected of containing traps.
    identify_traps_kwargs:
        Passed to identify_trap_locations

    Returns
    -------
    traps: an array of pairs of integers
        The coordinates of the centroids of the traps
    """
    # keep a memory of image in case need to re-run
    img = image
    # bounds on major axis length of traps
    min_mal = min_frac_tilesize * np.sqrt(2) * tile_size
    max_mal = max_frac_tilesize * np.sqrt(2) * tile_size

    # shrink image
    if downscale != 1:
        img = transform.rescale(image, downscale)
    # generate an entropy image using a disk footprint
    disk_radius = int(min([disk_radius_frac * x for x in img.shape]))
    entropy_image = entropy(img_as_ubyte(img), disk(disk_radius))
    if downscale != 1:
        entropy_image = transform.rescale(entropy_image, 1 / downscale)
    # find Otsu threshold for entropy image
    thresh = threshold_otsu(entropy_image)
    # apply morphological closing to thresholded, and so binary, image
    bw = closing(entropy_image > thresh, square(square_size))
    # remove artifacts connected to image border
    cleared = clear_border(bw)

    # label distinct regions of the image
    label_image = label(cleared)
    # find regions likely to contain traps:
    # with a major axis length within a certain range
    # and a centroid at least tile_size // 2 away from the image edge
    idx_valid_region = [
        (i, region)
        for i, region in enumerate(regionprops(label_image))
        if min_mal < region.major_axis_length < max_mal
        and tile_size // 2
        < region.centroid[0]
        < half_floor(image.shape[0], tile_size) - 1
        and tile_size // 2
        < region.centroid[1]
        < half_floor(image.shape[1], tile_size) - 1
    ]
    idx, valid_region = zip(*idx_valid_region)

    # find centroids and minor axes lengths of valid regions
    centroids = (
        np.array([x.centroid for x in valid_region]).round().astype(int)
    )
    minals = [region.minor_axis_length for region in valid_region]
    # coords for best trap
    x, y = np.round(centroids[np.argmin(minals)]).astype(int)

    # make a template using the best trap in the image
    template = image[
        half_floor(x, tile_size):half_ceil(x, tile_size),
        half_floor(y, tile_size):half_ceil(y, tile_size),
    ]
    # make candidate templates from the other traps found
    candidate_templates = [
        image[
            half_floor(x, tile_size):half_ceil(x, tile_size),
            half_floor(y, tile_size):half_ceil(y, tile_size),
        ]
        for x, y in centroids
    ]
    # make a mean template from all the found traps
    mean_template = np.dstack(candidate_templates).astype(int).mean(axis=-1)

    # find traps using the best found trap
    traps = identify_trap_locations(image, template, **identify_traps_kwargs)
    # find traps using the mean trap template
    mean_traps = identify_trap_locations(
        image, mean_template, **identify_traps_kwargs
    )
    # choose the approach that identifies the most traps
    if len(traps) < len(mean_traps):
        traps = mean_traps

    # if there are too few traps, try again
    traps_retry = []
    if len(traps) < 30 and downscale != 1:
        print("Tiler:TrapIdentification: Trying again.")
        traps_retry = segment_traps(image, tile_size, downscale=1)

    # return results with the most number of traps
    if len(traps_retry) < len(traps):
        return traps
    else:
        return traps_retry


###


def identify_trap_locations(
    image, trap_template, optimize_scale=True, downscale=0.35, trap_size=None
):
    """
    Identify the traps in a single image based on a trap template,
    which requires the trap template to be similar to the image
    (same camera, same magification - ideally the same experiment).

    Uses normalised correlation in scikit-image's match_template.

    The search is speeded up by downscaling both the image and
    the trap template before running the template matching.

    The trap template is rotated and re-scaled to improve matching.
    The parameters of the rotation and rescaling are optimised, although
    over restricted ranges.

    Parameters
    ----------
    image: 2D array
    trap_template: 2D array
    optimize_scale : boolean (optional)
    downscale: float (optional)
        Fraction by which to downscale to increase speed
    trap_size: integer (optional)
        If unspecified, the size is determined from the trap_template

    Returns
    -------
    coordinates: an array of pairs of integers
    """
    if trap_size is None:
        trap_size = trap_template.shape[0]
    # careful: the image is float16!
    img = transform.rescale(image.astype(float), downscale)
    template = transform.rescale(trap_template, downscale)

    # try multiple rotations of template to determine
    # which best matches the image
    # result is squared because the sign of the correlation is unimportant
    matches = {
        rotation: feature.match_template(
            img,
            transform.rotate(template, rotation, cval=np.median(img)),
            pad_input=True,
            mode="median",
        )
        ** 2
        for rotation in [0, 90, 180, 270]
    }
    # find best rotation
    best_rotation = max(matches, key=lambda x: np.percentile(matches[x], 99.9))
    # rotate template by best rotation
    template = transform.rotate(template, best_rotation, cval=np.median(img))

    if optimize_scale:
        # try multiple scales appled to template to determine which
        # best matches the image
        scales = np.linspace(0.5, 2, 10)
        matches = {
            scale: feature.match_template(
                img,
                transform.rescale(template, scale),
                mode="median",
                pad_input=True,
            )
            ** 2
            for scale in scales
        }
        # find best scale
        best_scale = max(
            matches, key=lambda x: np.percentile(matches[x], 99.9)
        )
        # choose the best result - an image of normalised correlations
        # with the template
        matched = matches[best_scale]
    else:
        # find the image of normalised correlations with the template
        matched = feature.match_template(
            img, template, pad_input=True, mode="median"
        )
    # re-scale back the image of normalised correlations
    # find the coordinates of local maxima
    coordinates = feature.peak_local_max(
        transform.rescale(matched, 1 / downscale),
        min_distance=int(trap_size * 0.70),
        exclude_border=(trap_size // 3),
    )
    return coordinates


###############################################################
# functions below here do not appear to be used any more
###############################################################


def stretch_image(image):
    image = ((image - image.min()) / (image.max() - image.min())) * 255
    minval = np.percentile(image, 2)
    maxval = np.percentile(image, 98)
    image = np.clip(image, minval, maxval)
    image = (image - minval) / (maxval - minval)
    return image


def get_tile_shapes(x, tile_size, max_shape):
    half_size = tile_size // 2
    xmin = int(x[0] - half_size)
    ymin = max(0, int(x[1] - half_size))
    # if xmin + tile_size > max_shape[0]:
    #     xmin = max_shape[0] - tile_size
    # if ymin + tile_size > max_shape[1]:
    # #     ymin = max_shape[1] - tile_size
    # return max(xmin, 0), xmin + tile_size, max(ymin, 0), ymin + tile_size
    return xmin, xmin + tile_size, ymin, ymin + tile_size


def in_image(img, xmin, xmax, ymin, ymax, xidx=2, yidx=3):
    if xmin >= 0 and ymin >= 0:
        if xmax < img.shape[xidx] and ymax < img.shape[yidx]:
            return True
    else:
        return False


def get_xy_tile(img, xmin, xmax, ymin, ymax, xidx=2, yidx=3, pad_val=None):
    if pad_val is None:
        pad_val = np.median(img)
    # Get the tile from the image
    idx = [slice(None)] * len(img.shape)
    idx[xidx] = slice(max(0, xmin), min(xmax, img.shape[xidx]))
    idx[yidx] = slice(max(0, ymin), min(ymax, img.shape[yidx]))
    tile = img[tuple(idx)]
    # Check if the tile is in the image
    if in_image(img, xmin, xmax, ymin, ymax, xidx, yidx):
        return tile
    else:
        # Add padding
        pad_shape = [(0, 0)] * len(img.shape)
        pad_shape[xidx] = (max(-xmin, 0), max(xmax - img.shape[xidx], 0))
        pad_shape[yidx] = (max(-ymin, 0), max(ymax - img.shape[yidx], 0))
        tile = np.pad(tile, pad_shape, constant_values=pad_val)
    return tile


def tile_where(centre, x, y, MAX_X, MAX_Y):
    # Find the position of the tile
    xmin = int(centre[1] - x // 2)
    ymin = int(centre[0] - y // 2)
    xmax = xmin + x
    ymax = ymin + y
    # What do we actually have available?
    r_xmin = max(0, xmin)
    r_xmax = min(MAX_X, xmax)
    r_ymin = max(0, ymin)
    r_ymax = min(MAX_Y, ymax)
    return xmin, ymin, xmax, ymax, r_xmin, r_ymin, r_xmax, r_ymax


def get_tile(shape, center, raw_expt, ch, t, z):
    """Returns a tile from the raw experiment with a given shape.

    :param shape: The shape of the tile in (C, T, Z, Y, X) order.
    :param center: The x,y position of the centre of the tile
    :param
    """
    _, _, x, y, _ = shape
    _, _, MAX_X, MAX_Y, _ = raw_expt.shape
    tile = np.full(shape, np.nan)

    # Find the position of the tile
    xmin = int(center[1] - x // 2)
    ymin = int(center[0] - y // 2)
    xmax = xmin + x
    ymax = ymin + y
    # What do we actually have available?
    r_xmin = max(0, xmin)
    r_xmax = min(MAX_X, xmax)
    r_ymin = max(0, ymin)
    r_ymax = min(MAX_Y, ymax)

    # Fill values
    tile[
        :,
        :,
        (r_xmin - xmin): (r_xmax - xmin),
        (r_ymin - ymin): (r_ymax - ymin),
        :,
    ] = raw_expt[ch, t, r_xmin:r_xmax, r_ymin:r_ymax, z]
    # fill_val = np.nanmedian(tile)
    # np.nan_to_num(tile, nan=fill_val, copy=False)
    return tile


def get_traps_timepoint(
    raw_expt, trap_locations, tp, tile_size=96, channels=None, z=None
):
    """
    Get all the traps from a given time point
    :param raw_expt:
    :param trap_locations:
    :param tp:
    :param tile_size:
    :param channels:
    :param z:
    :return: A numpy array with the traps in the (trap, C, T, X, Y,
    Z) order
    """

    # Set the defaults (list is mutable)
    channels = channels if channels is not None else [0]
    z_positions = z if z is not None else [0]
    if isinstance(z_positions, slice):
        n_z = z_positions.stop
        z_positions = list(range(n_z))  # slice is not iterable error
    elif isinstance(z_positions, list):
        n_z = len(z_positions)
    else:
        n_z = 1

    n_traps = len(trap_locations[tp])
    trap_ids = list(range(n_traps))
    shape = (len(channels), 1, tile_size, tile_size, n_z)
    # all tiles
    zct_tiles, slices, trap_ids = all_tiles(
        trap_locations, shape, raw_expt, z_positions, channels, [tp], trap_ids
    )
    # TODO Make this an explicit function in TimelapseOMERO
    images = raw_expt.pixels.getTiles(zct_tiles)
    # Initialise empty traps
    traps = np.full((n_traps,) + shape, np.nan)
    for trap_id, (z, c, _, _), (y, x), image in zip(
        trap_ids, zct_tiles, slices, images
    ):
        ch = channels.index(c)
        z_pos = z_positions.index(z)
        traps[trap_id, ch, 0, x[0]: x[1], y[0]: y[1], z_pos] = image
    for x in traps:  # By channel
        np.nan_to_num(x, nan=np.nanmedian(x), copy=False)
    return traps


def centre(img, percentage=0.3):
    y, x = img.shape
    cropx = int(np.ceil(x * percentage))
    cropy = int(np.ceil(y * percentage))
    startx = int(x // 2 - (cropx // 2))
    starty = int(y // 2 - (cropy // 2))
    return img[starty: starty + cropy, startx: startx + cropx]


def align_timelapse_images(
    raw_data, channel=0, reference_reset_time=80, reference_reset_drift=25
):
    """
    Uses image registration to align images in the timelapse.
    Uses the channel with id `channel` to perform the registration.

    Starts with the first timepoint as a reference and changes the
    reference to the current timepoint if either the images have moved
    by half of a trap width or `reference_reset_time` has been reached.

    Sets `self.drift`, a 3D numpy array with shape (t, drift_x, drift_y).
    We assume no drift occurs in the z-direction.

    :param reference_reset_drift: Upper bound on the allowed drift before
    resetting the reference image.
    :param reference_reset_time: Upper bound on number of time points to
    register before resetting the reference image.
    :param channel: index of the channel to use for image registration.
    """
    ref = centre(np.squeeze(raw_data[channel, 0, :, :, 0]))
    size_t = raw_data.shape[1]

    drift = [np.array([0, 0])]
    for i in range(1, size_t):
        img = centre(np.squeeze(raw_data[channel, i, :, :, 0]))

        shifts, _, _ = phase_cross_correlation(ref, img)
        # If a huge move is detected at a single time point it is taken
        # to be inaccurate and the correction from the previous time point
        # is used.
        # This might be common if there is a focus loss for example.
        if any(
            [
                abs(x - y) > reference_reset_drift
                for x, y in zip(shifts, drift[-1])
            ]
        ):
            shifts = drift[-1]

        drift.append(shifts)
        ref = img

        # TODO test necessity for references, description below
        #   If the images have drifted too far from the reference or too
        #   much time has passed we change the reference and keep track of
        #   which images are kept as references
    return np.stack(drift), ref
